#-*- coding: ISO-8859-15 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------
'''
Created on 20 fvr. 2017

@author: jeanluc
'''
import os, sys, stat
import functools as ft

#from colorama import Fore, Style
from sqlite3 import connect as sqlconnect, \
    PARSE_DECLTYPES as sqlPARSE_DECLTYPES, \
    OperationalError as sqlOperationalError
#     Connection as sqlConnection, \
from argparse import ArgumentParser, FileType
from collections import defaultdict, OrderedDict
from six import text_type
import numpy as np
import pandas, openpyxl, xlrd
from pandas import DataFrame

import nntoolbox.utils as nnu
from nntoolbox import __version__ as nntoolboxversion
from nntoolbox.utils import doNothing, getListFromStr, isIntList,\
    splitvarunit, float_format, IDformat, roundoff_format, defaultvalue,\
    isIntBase, isInt, unquote, callerView
from nntoolbox.modelutils import getModelDict, getGmxFileAndSave
from nntoolbox.excelutils import multiRangesType, simpleRangeType, \
    getSheet  #outputmarker, 
from nntoolbox.filetoolbox import getReducedSelectedGmx, \
    decorfile, arg_file_parse, option2Arguments, IndexFilename
from nntoolbox.progress import FanWait, Progress_tqdm
from nntoolbox.configparsertoolbox import defaultedConfigParser, \
    getDefaultedConfigParser 
from nntoolbox.sqlite_numpy import createTableDb, cleanTableDb, stdTable
from nntoolbox.utils import finished  

from monal import __version__ as monalversion
from monal import monalconst as C
from monal.driver import onEndIter, onReturnResult, Driver  #DriverLib, 
from monal.util.toolbox import setTrainingActionParams, orderedIndexes
from monal.util.toolbox import toPrint
from monal.util.lineartoolbox import randomProbe  
from monal.util.monaltoolbox import choiceQuestion, valueQuestion, \
    yesNoQuestion, chooseFields, debugOutput
from monal.datareader.excelsource import gettitles
from monal.datareader.datasource import DataSource
from monal.modelutils import loadModule
from monal.util.utils import CCharOnlys, getapplidatabase
from monal.datareader.excelsource import getcustomproperty 
from monal.datareader.modeldataframe import get_modelDataframe
from monal.monalrecords import ModelData

from chem_gm.core.gmprograms import analysebase
from chem_gm.core.gmutils import getConfigString, ReverseConfigString, \
    key2long, cyclekey
from chem_gm.core.gmconst import defaultConnect
from chem_gm.core.gmtoolbox import fromConfigstr2Dict
from chem_gm.components.gmmEngine import createGMModel #, computeSmile

from nn1 import __version__ as nn1version
from nn1.nn1_api_service import _CheckTrainingParameters, _checkGroupList, \
    _getCompiledModuleNN, _checkFile, getResultFrame, apiTrainResults, \
    apiTestResults, _CheckLOOParameters, _CheckBSParameters, \
    _apiDemoBS, _apiDemoLOO, printresult, basicDecorDict, \
    createTrainingModule, defaultCfg, LoadModuleFromCfg, \
    getDataFileInDocker, prepareUsageAction, finishSingleUsage, \
    manageSingleUsage, getSimpleExtraFields, finishMultiUsage, \
    createExcelwriter, CloseExcelwriter, updateOptionsFromGmx, \
    _getBestWeights, internalshare, internalresult, internaldata, \
    internalargsdir, nobarverb, currentshare  #defaultworkingdir,

from chem_gm import __version__ as gmversion

class nn1Error(Exception):
    pass

dicoversion = OrderedDict()
dicoversion['nn1'] = nn1version
dicoversion['chem_gm'] = gmversion
dicoversion['monal'] = monalversion
dicoversion['nntoolbox'] = nntoolboxversion
dicoversion['numpy'] = np.__version__
dicoversion['pandas'] = pandas.__version__
dicoversion['openpyxl'] = openpyxl.__version__
dicoversion['xlrd'] = xlrd.__VERSION__
  

applidatapath = getapplidatabase("nn1")

preproctypelist = ['', 'linear', 'simple2ndorder', 'full2ndorder']
""" simple2ndorder -> no cross products in the second order.
    full2ndorder -> full second order.
"""
preprocnamelist = ['', 'linear', 'simple quadratic', 'full quadratic']

_argumentsdefaultdict = {
    "HELP": 0,
    "VERSION": 0,
    "LICENSE": 0,
    "YES": 1,
    "YESYES": 0,
    "ACTION": 'analyse,preproc',
    "GROUPLIST": "",  #"id,in,out",
    "GROUPCOUNT": "0",
#     SHAREFOLDER=/Users/jeanluc/nn1
    "SORTBY": 'PRESS',
    "RESULTTYPE": 'xlsx',
    "MAXREC": "10:20:5",
    "HIDDEN": 5,
    'CONFIGSTR': '',
    'CENTRAL': '',
    'ORIGIN': '',
    'SAVEFOLDER': '',
    'CLASSIF': 0,
    "EPOCHS": 150,
    "INIT": 150,
    "PROBE": 1000,
    "SEED": -1,
    "DECOR": "t,i,e,s,se,lt",
    "LOOINITTYPE": 2,
    "LOOINITCOUNT": 100,
    "LOOEPOCHS": 150,
    "LOOMAXREC": 1,
    "LOOLEVTHRES": 5.0,
    "BSTYPE": 2,
    "BSINITTYPE": 2,
    "BSCOUNT": 250,
    "BSINITCOUNT": 2,
    "BSEPOCHS": 150,
    "PREPROCTYPE": "",
    "PERTTHRES": 0.5,
    "RDMINIT": 0.1,
    "RNDTHRES": 1E-4,
    "FORCEMODEL": 0,
    "VERB": 3,
    "CALLER": 0,
    'LIBUSE': 0,
    'TIMER': 0,
#    "DEBUG": 0,
    }

groupdict = {
    'inputs':'inputs',
    'in': 'inputs',
    'outputs': 'outputs',
    'out': 'outputs',
    'identifiers':'identifiers',
    'id':'identifiers',
    'smiles':'smiles',
    'smi':'smiles',
    }

def _getCmdLineOptions(args, defaultdico=None):
    """Read the argument list, usually coming from command line.
    defaultdico is the dictionary of default values.
    
    Return the 'options' namespace.
    """
    #if defaultdico is None:
    origindico = defaultdict(lambda: None)
    origindico.update(_argumentsdefaultdict)
    if defaultdico is not None:
        origindico.update(defaultdico) 
    defaultdico = origindico
    actionlist = ['all', 'analyse', 'preproc', 'train', 'test', 'loo', 'lootest', 'bootstrap', 'bootstraptest', 'usage']
    actionlist.extend([val.upper() for val in actionlist])
    fullpreproctypelist = preproctypelist + list(range(len(preproctypelist)))
    parser = ArgumentParser(description='nn1 API demo.', fromfile_prefix_chars='@') 
    # string arguments
    parser.add_argument('--arg_file', nargs="?", type=str, default="",
        help="name of arguments file")
    parser.add_argument('source', nargs="?", type=str,
        default=defaultvalue("DATAFILE", "", "", defaultdico=defaultdico),
        help="Initial data source filename (xls,xlsx,xlsm,csv,txt format)")
    parser.add_argument('usagedata', nargs="?", type=str,
        default=defaultvalue("DATA", "", "", defaultdico=defaultdico),
        help="data for single usage")
    parser.add_argument('test', nargs="?", type=str,
        default=defaultvalue("TESTFILE", "", "", defaultdico=defaultdico),
        help="Data test filename (xls,xlsx,csv,txt format)")
    parser.add_argument('-typ', '--filetype', default="",
        choices=["txt", "csv", "xls", "xlsx"], nargs='?',
        help="Source and test file type")
    parser.add_argument('-cc', '--compiler', dest='compiler', 
        default="", help="Compiler for model libraries")
    parser.add_argument('-sf', '--savefolder', nargs='?', 
        default=defaultvalue("SAVEFOLDER", "", "", defaultdico=defaultdico),
        help="training results saving folder")
    parser.add_argument('-or', '--origin', nargs='?', 
        default=defaultvalue("ORIGIN", "", "", defaultdico=defaultdico),
        help="training results saving folder")
    parser.add_argument('-f', '--folder', dest='folder', 
        default=applidatapath, help="Set working folder")
    parser.add_argument('-sd', '--datarange', 
        default=defaultvalue("DATARANGE", "DATA", "", defaultdico=defaultdico),
        help="Range for training data (Excel-like files)")
    parser.add_argument('-st', '--testrange', 
        default=defaultvalue("TESTRANGE", "TEST", "", defaultdico=defaultdico),
        help="Range for test data (Excel-like files)")
    parser.add_argument('-fd', '--datafields', 
        default=defaultvalue("DATAFIELDS", "", "", defaultdico=defaultdico),            
        help="Fields for training data (Excel-like files)")
    parser.add_argument('-ft', '--testfields',
        default=defaultvalue("TESTFIELDS", "", "", defaultdico=defaultdico),            
        help="Fields for test data (Excel-like files)")
    parser.add_argument('-dt', '--datatable', default=stdTable, 
        help="SQL table name")
    parser.add_argument('-C', '--configstr', 
        default=defaultvalue("CONFIGSTR", "", "", defaultdico=defaultdico), 
        nargs="?", help="Atoms and structures connectivity string")
    parser.add_argument('-gl', '--grouplist', 
        default = defaultvalue("GROUPLIST", "", "", defaultdico=defaultdico),
        help='field groups')
    parser.add_argument('-gc', '--groupcount', 
        default=defaultvalue("GROUPCOUNT", "0", "0", defaultdico=defaultdico),
        help='field counts in groups')
    parser.add_argument('-u', '--usesharedfolder', nargs="?", const="", 
        type=str, default=defaultvalue("SHAREFOLDER", "", "", "", defaultdico=defaultdico), 
        help="share working folder with host")
    parser.add_argument('-r', '--resultindex', type=int, 
        default=defaultvalue("RESINDEX", "0", "0", defaultdico=defaultdico),
        help="Best result index. Default 0")
    parser.add_argument('-k', '--keeptemp', dest='keeptemp', 
        default=defaultvalue("TEMPDIR", "", "", defaultdico=defaultdico),
        help="Keeping temporary folder")
    parser.add_argument('-o', '--orderfield', const="VLOO score", nargs="?", 
        default=defaultvalue("SORTBY", default="", const="VLOO score", defaultdico=defaultdico),
        help="Data field used for result sorting")
    parser.add_argument('-d', '--decor', nargs='?', 
        default=defaultvalue("DECOR", '', '', defaultdico=defaultdico),
        help="decorator rules for result files. Default ''")
    parser.add_argument('-pu', '--propertyunit', 
        default = defaultvalue("PROPERTYUNIT", "", "", defaultdico=defaultdico),
        help="unit of the targeted property")
    
    # integer arguments
    parser.add_argument('--license', nargs="?", const=1, type=int, 
        default=defaultvalue("LICENSE", 0, 0, defaultdico=defaultdico),
        help="Display license")
    parser.add_argument('--version', nargs="?", const=1, type=int, 
        default=defaultvalue("VERSION", 0, 0, defaultdico=defaultdico),
        help="Display graphmachine docker container version")
    parser.add_argument('--dhelp', nargs="?", const=1, type=int, 
        default=defaultvalue("HELP", 0, 0, defaultdico=defaultdico),
        help="Display docker container help")
    parser.add_argument('-y', '--yes', nargs="?", const=1, type=int,
        default=defaultvalue("YES", defaultdico=defaultdico),
        help="Answer yes to all questions and accept default values")
#     parser.add_argument('-yy', '--yesyes', nargs="?", const=0, type=int,
#         default=defaultvalue("YESYES", defaultdico=defaultdico),
#         help="Answer yes to all questions in create arg file and accept default values")
    parser.add_argument('-v', '--verbose', 
        default=defaultvalue("VERB", "1", "1", defaultdico=defaultdico), 
        type=str, help="Verbose level")
    parser.add_argument('-R', '--copytype', 
        default=defaultvalue("RESULTTYPE", "xlsx", "xlsx", defaultdico=defaultdico),
        help="result copy file type")
    parser.add_argument('-M', '--maxrec', 
        default=defaultvalue("MAXREC", 10, 0, defaultdico=defaultdico),
        help="Number max of registered records (list of); default 10")
    
    parser.add_argument('-s', '--seed', 
        default=defaultvalue("SEED", -1, -1, defaultdico=defaultdico), 
        help="Random generator seed. Default -1")
    
    parser.add_argument('-H', '--hidden', 
        default=defaultvalue("HIDDEN", 5, 0, defaultdico=defaultdico),
        help="Number of hidden neurons (list of). Default 5")
    parser.add_argument('-e', '--epochs', type=int, 
        default=defaultvalue("EPOCHS", 250, 0, outtype=int, defaultdico=defaultdico), 
        help="Number of training epochs. Default 250")
    parser.add_argument('-I', '--initcount', type=int, 
        default=defaultvalue("INIT", 100, 0, outtype=int, defaultdico=defaultdico), 
        help="Number of training initialization. Default 100")
    parser.add_argument('-pc', '--probecount', type=int, 
        default=defaultvalue("PROBE", 100, 0, defaultdico=defaultdico), 
        help="Number of random probe initialization. Default 100")
    parser.add_argument('--caller', type=int, 
        default=defaultvalue("CALLER", 0, 0, defaultdico=defaultdico),
        help="Caller type (0: direct, 1: docker). Default 0")
    parser.add_argument('-rd', '--resultdisplay', type=int, default=-1, 
        help="Number of best result displayed. Default -1 (=all)")
    
    parser.add_argument('-lt', '--looinittype', type=int, 
        default=defaultvalue("LOOINITTYPE", 2, 0, defaultdico=defaultdico), 
        help="special init 0: start weight, 1: end weight, 2 random). Default 0")
    parser.add_argument('-li', '--looinitcount', type=int, 
        default=defaultvalue("LOOINITCOUNT", 100, 0, defaultdico=defaultdico), 
        help="LOO: number of train per base. Default 100")
    parser.add_argument('-le', '--looepochs', type=int, 
        default=defaultvalue("LOOEPOCHS", 150, 0, defaultdico=defaultdico), 
        help="number of epochs in LOO trainings. Default 150")
    parser.add_argument('-lm', '--loomaxrec', type=int, 
        default=defaultvalue("LOOMAXREC", 1, 0, defaultdico=defaultdico), 
        help="number of training selected after LOO trainings. Default 1")
    
    parser.add_argument('-bs', '--bootstraptype', type=int, 
        default=defaultvalue("BSTYPE", 2, 0, defaultdico=defaultdico), 
        help="bootstraptype 0: none, 1: standard, 2: residus. Default 2")
    parser.add_argument('-bt', '--bootstrapinittype', type=int, 
        default=defaultvalue("BSINITTYPE", 2, 0, defaultdico=defaultdico), 
        help="special init 0: start weight, 1: end weight, 2 random). Default 2")
    parser.add_argument('-bc', '--bootstrapcount', type=int, 
        default=defaultvalue("BSCOUNT", 250, 0, defaultdico=defaultdico), 
        help="boostrap: number of train per base. Default 100")
    parser.add_argument('-bi', '--bootstrapinitcount', type=int, 
        default=defaultvalue("BSINITCOUNT", 100, 0, defaultdico=defaultdico),
        help="boostrap: number of train initialisation per base. Default 100")
    parser.add_argument('-be', '--bootstrapepochs', type=int, 
        default=defaultvalue("BSEPOCHS", 150, 0, defaultdico=defaultdico), 
        help="number of epochs in bootstrap trainings. Default 150")
    
    # choices arguments
    parser.add_argument('-a', '--actionstr', dest='actionstr', 
        default=defaultvalue("ACTION", "", "", defaultdico=defaultdico),
        help="action option in character string version with comma separator")
    
    parser.add_argument('-pp', '--preproctype', choices=fullpreproctypelist, 
        default=defaultvalue("PREPROCTYPE", "linear", "linear", 
                             defaultdico=defaultdico),  
        help="""Preproc type: 
            '' -> none, 
            'repartition' -> random probe linear function', 
            'normal -> random probe with limited 2nd order' 
            'full -> random probe with 2nd order'""")
    # float arguments
    parser.add_argument('-D', '--initstddev', type=float,
        default=defaultvalue("RDMINIT", 0.1, 0, outtype=float, 
                             defaultdico=defaultdico),
        help="Standard deviation for param initialization. Default 0.1")  
    parser.add_argument('-pt', '--pertinencethreshold', type=float,
        default=defaultvalue("PERTTHRES", 0.5, 0, outtype=float, 
                             defaultdico=defaultdico),
        help="Pertinence threshold for input selection")
    parser.add_argument('-T', '--roundoffthreshold', type=float, 
        default=defaultvalue("RNDTHRES", 1E-4, 0, outtype=float, 
                             defaultdico=defaultdico),
        dest="roundthreshold", help="Roundoff selection threshold. Default 1E-4")
    parser.add_argument('-L', '--leveragethreshold', type=float, 
        default=defaultvalue("LEVTHRES", 5.0, 0, outtype=float, 
                             defaultdico=defaultdico),
        help="Roundoff selection threshold. Default 5.0")
    parser.add_argument('-LL', '--looleveragethreshold', type=float, 
        default=defaultvalue("LOOLEVTHRES", 1.0, 0, outtype=float, 
                             defaultdico=defaultdico),
#         default=float(os.getenv("LOOLEVTHRES", 5.0)),
        help="Roundoff selection threshold for LOO. Default 5.0")  
    
    # boolean arguments  
    parser.add_argument('-cl', '--classif', action='store_true',# nargs="?", const=1,
        default=defaultvalue("CLASSIF", "", "0", defaultdico=defaultdico),
        help="Create a classification model")
    parser.add_argument('-m', '--modelcreation', action='store_true',# nargs="?", const=1,
        default=defaultvalue("FORCEMODEL", "", "0", defaultdico=defaultdico),
        help="No model creation if already exists")
    parser.add_argument('--timer', action='store_true', 
        default=defaultvalue("TIMER", 0, 0, defaultdico=defaultdico), 
        dest="dotimer", help="Set timer mode. Default 0")
    parser.add_argument('--debug', action='store_true', 
        default=defaultvalue("DEBUG", 0, 0, defaultdico=defaultdico), 
        help="Debug state. Default 0")
    parser.add_argument('-lu', '--libuse', action='store_true', 
        default=defaultvalue("LIBUSE", 0, 0, defaultdico=defaultdico), 
        help="Debug state. Default 0")
    # file arguments
    parser.add_argument('--infile', type=FileType('r'), 
        default=sys.stdin, help='Input file. Default stdin')
    parser.add_argument('--outfile', type=FileType('w'),
        default=sys.stdout, help='Output file. Default stdout')
     
    return parser.parse_args(args)

def _hasTrain(options):
    return 'train' in options.actionlist or 'loo' in options.actionlist or 'bootstrap' in options.actionlist

def _hasTest(options):
    return 'test' in options.actionlist or 'lootest' in options.actionlist or 'bootstraptest' in options.actionlist #or 'loo' in options.actionlist or 'bootstrap' in options.actionlist

def _actionListMgmt(llist):
    if llist == ['all', 'loo']:
        llist = ['train', 'test', 'loo']
    elif llist == ['all', 'bootstrap']:
        llist = ['train', 'test', 'bootstrap']
    elif llist == ['all']:
        llist = ['train', 'test']
#     if ('loo' in llist) and not ('lootest' in llist):
#         llist.append('lootest')
    return llist

def _getDockerDataFile(filename, caller=0, searchdir="", candidates=[], 
        datakind="datafile", defaultindex=None, forcedefault=False, 
        doprint=doNothing):
    index = None
    if not searchdir:
        searchdir = os.path.join(currentshare(caller), "data")
    if not filename:
        try:
            if not len(candidates):
                candidates, searchdir = _getDataFileCandidates(searchdir=searchdir)
            if len(candidates):
                mess = "Search folder : {0}\n".format(searchdir)
                if datakind:
                    mess += "Will you choose the {0} in the candidate list ?".format(datakind)                    
                    printMessage = "chosen {0} ".format(datakind) 
                else:
                    mess += "Will you choose the datafile in the candidate list ?"
                    printMessage = "chosen datafile "
                #mess = ""
                res, index = choiceQuestion(mess, candidates,
                    printMessage=printMessage, 
                    returnType=str, listindexed=True, extradisplay=True,
                    defaultindex=defaultindex, fulloutput=True, 
                    forcedefault=forcedefault, doprint=doprint)
                if res:
                    candidate = os.path.join(searchdir, res)
                    if os.path.exists(candidate):
                        filename = candidate
        except: pass
    if not filename:
        return "", -1
    if os.path.exists(filename):
        return filename, index
    test = os.path.join(searchdir, filename)    
    if os.path.exists(test):  
        return test
    ppath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    pathlst = ["/data",
               os.path.join(ppath, "api", "demofiles"),
               os.path.join(applidatapath, "demofiles"),
               ]
    for curpath in pathlst:
        test = os.path.join(curpath, filename)    
        if os.path.exists(test):  
            return test, index
    return "", -1  
    
def _isSourceFile(filename, exts=[]):
    _, ext = os.path.splitext(filename)
    if len(exts):
        OK = ext in exts
    else:
        OK = ext in [".txt", ".csv"] or (ext.startswith(".xls"))
    return OK
    
def _getDataFileCandidates(searchdir="", exts=[], searchbase=[]):
    candidates = []
    for sdir in searchbase:
        try:
            _, _, files = next(os.walk(sdir))
            candidates = [fil for fil in files if _isSourceFile(fil, exts)] 
        except: pass
        if len(candidates):
            break
    if not len(candidates) and searchdir and os.path.isdir(searchdir): 
        try:
            _, _, files = next(os.walk(searchdir))
            candidates = [fil for fil in files if _isSourceFile(fil, exts)] 
        except: pass
    return candidates, searchdir

def checkInDir(filename, dirname):
    test = os.path.join(dirname, filename)
    if os.path.exists(test):
        return True, test
    for root, dirs, _ in os.walk(dirname):
        for fold in dirs:
            found, filename = checkInDir(filename, os.path.join(root, fold)) 
            if found:
                return found, filename
    return False, filename
    
def _setFileFolders(filename, sharebaselist=[], externbase=""):    
    filename = unquote(filename)
    if not filename:
        return ""
    filename = os.path.expanduser(filename)
    if externbase and filename.startswith(externbase):
        # le nom du fichier est en reference externe: on 
        filename = filename[len(externbase):]
        if filename.startswith('/') or filename.startswith('\\'):
            filename = filename[1:]
    for basedir in sharebaselist:
        test = os.path.join(basedir, filename)
        if os.path.exists(test):
            return test
        for root, dirs, _ in os.walk(basedir):           
            test = os.path.join(root, filename)
            if os.path.exists(test):
                return test
    return filename

def _initiate_options(argv=[], proposedOptions=None, help=False):
    if help:
        argv.append("--help")
    dico = None
    argumentfile = ""
    dict, argumentfile = arg_file_parse(argv, ("@", "--arg_file"), full=True)        
    if isinstance(dict, text_type):
        sys.stderr.write(dict)
    elif dict is not None:
        dico = defaultdict(lambda: None)
        dico.update(dict)
    if proposedOptions:
        options = proposedOptions
    else:
        options = _getCmdLineOptions(argv, dico)
    if int(options.yes) & 2:
        options.yesyes = True
        options.yes = options.yesyes % 2
    else:
        options.yesyes = False
#     options.internalyes = (int(options.yes) < 0)
#     if options.internalyes:
#         options.yes = _argumentsdefaultdict["YES"]
        
    options.argumentfile = argumentfile
    return options

def _initiate_debug(options):
    toprint = ""
    errorcode = nnu.OPTION_NO_ERROR
    debugbase = isIntBase(options.debug) # return 0, 8, 10 ou 16
    if debugbase:
        # ici le debug est appele par son code en entier, octal ou hexadecimal
        options.debug = int(options.debug, debugbase)
    elif options.debug and isinstance(options.debug, text_type):
        # ici le debug est appele par une liste de 
        lst = options.debug.lower().split(',')
        value = 0
        for test in lst:
            if test == 'options':
                value |= nnu.DEMO_DEBUG_OPTIONS
            elif test == 'analysis':
                value |= nnu.DEMO_DEBUG_ANALYSIS
            elif test == 'model':
                value |= nnu.DEMO_DEBUG_MODEL
            elif test == 'transfer':
                value |= nnu.DEMO_DEBUG_TRANSFER
            elif test == 'jacob':
                value |= nnu.DEMO_DEBUG_JACOBIAN
            elif test == 'cost':
                value |= nnu.DEMO_DEBUG_COST
            elif test == 'press':
                value |= nnu.DEMO_DEBUG_PRESS
            elif test == 'parallel':
                value |= nnu.DEMO_DEBUG_PARALLEL  
            elif test == 'usage':
                value |= nnu.DEMO_DEBUG_USAGE
            elif test == 'folder':
                value |= nnu.DEMO_DEBUG_FOLDER
            else:
                toprint = 'unknown debug option : "%s"\n'% test 
                errorcode |= nnu.OPTION_WARN_AND_STOP
                break
        options.debug = value
    return options, toprint, errorcode

def initiate_verbose(options):
    verb = [int(val) for val in str(options.verbose).split(',')]
#     options.dotimer = False
    if len(verb) == 1:
        options.verbose = int(verb[0]) % 16
        options.verbosexls = int(verb[0]) // 16
    else:
        options.verbose = int(verb[0])
        options.verbosexls = int(verb[1])
        if len(verb) > 2:
            options.dotimer = int(verb[2])
    return options

def localize_file(options, filename):
    if filename.startswith("/home") and options.externpath:
        return filename.replace("/home", options.externpath, 1)
    return filename

def initiate_externpath(options):
    createargfile = 'analyse' in options.actionlist or ('preproc' in options.actionlist and len(options.actionlist)==1)
    searchbase = currentshare(options.caller)
    options.externpath = ""
    pthfile = os.path.join(searchbase, 'host.pth')
    if os.path.exists(pthfile):
        with open(pthfile, "r") as ff:
            options.externpath = ff.read().strip()
    if not options.externpath and not options.caller:
        #if os.path.exists(options.usesharedfolder):
        options.externpath = options.usesharedfolder 

    options.indirectoutput = False
    if options.caller:
        # if options.usesharedfolder: 
        # cas du docker
        if not createargfile:
            options.usesharedfolder = internalshare
        options.indirectoutput = True
        options.folder = internalshare
    else:
        if not options.usesharedfolder:
            options.usesharedfolder = currentshare(options.caller)
            options.indirectoutput = True
        elif not os.path.exists(options.usesharedfolder):
            options.usesharedfolder = currentshare(options.caller) 
        options.folder = options.usesharedfolder
    
    
    return options

def api_options(argv=[], proposedOptions=None, nofile=False, help=False):  #callermodule="nn1", 
    """Options namespace creating or reading from command line,
    
    parameters:
     - argv            -> comand line argument list. Defaulted to empty list.
     - proposedOptions -> suggested option namespace. Defaulted to empty list.
     - searchdir       -> directory where data files are looked for when proposedOptions has no datafile. Defaulted to internal datadir
     - nofile          -> if set to True, the datafile searching is forbidden. Defaulted to False.
     - help            -> return a help text for the command line arguments. Defaulted to False.
     
    return: 
     - options         -> options namespace.
     - toprint         -> error message to print or text file to read and print when needed.
     - errorcode       -> error code. 
    Error codes are defined in 'nntoolbox.utils' 
     - OPTION_NO_ERROR = 0
     - OPTION_DISPLAY_FILE_AND_STOP = 1
     - OPTION_WARN_AND_STOP = 2
     - OPTION_WARN_AND_ASK = 4
     - OPTION_STOP = 5
    """

    toprint = ""
    errorcode = nnu.OPTION_NO_ERROR

# ici on recherche les options telles que lues dans la ligne de 
# commande ou les variables d'environnement
    options = _initiate_options(argv, proposedOptions, help)
    options.actionlist = options.actionstr.split(',')
#     createargfile = 'analyse' in options.actionlist or ('preproc' in options.actionlist and len(options.actionlist)==1)
# initialisations apres options primaires
    searchbase = currentshare(options.caller)
    searchbases = [searchbase]
    if not options.caller and options.usesharedfolder:
        searchbases.insert(0, options.usesharedfolder)
    searchdir = os.path.join(searchbase, "data")
    options.savedir = ""
    options.notest = False
    doprint = toPrint(options.outfile)
    
# traitement du debug 
    options, toprint, errorcode = _initiate_debug(options) 
    if  errorcode == nnu.OPTION_WARN_AND_STOP:
        return options, toprint, errorcode    
# adaptation des verboses
    options = initiate_verbose(options)
            
# recherche d'information sur le chemin externe (dans la machine hote)
    options = initiate_externpath(options)
            
# adaptations des chemeins des fichiers externes et internes           
#     sharebase = [searchbase]
#     if options.usesharedfolder and not options.caller:
#         sharebase.append(options.usesharedfolder)

    options.origin = _setFileFolders(options.origin, searchbases, options.externpath)
    options.source = _setFileFolders(options.source, searchbases, options.externpath)
    options.test = _setFileFolders(options.test, searchbases, options.externpath)
    if options.verbose > 4:    
        doprint("extern path", options.externpath)
        doprint("origin", options.origin)
        doprint("source", options.source)
        doprint("test", options.test)
    if not options.source and options.origin:
        options.actionstr = 'usage'

#     options.indirectoutput = False
#     if options.caller and options.usesharedfolder: 
#         # cas du docker
#         options.externpath = options.usesharedfolder
#         options.usesharedfolder = internalshare
#         options.indirectoutput = True
#     elif not options.usesharedfolder:
#         options.usesharedfolder = currentshare(options.caller)
#         options.indirectoutput = True
#     elif not os.path.exists(options.usesharedfolder):
#         options.usesharedfolder = currentshare(options.caller) 
            
#     if options.usesharedfolder:
#         if options.source and not os.path.exists(options.source):
#             temp = os.path.join(options.usesharedfolder, "data", options.source)
#             if os.path.exists(temp):
#                 options.source = temp
#             else:
#                 temp = getDataFileInDocker(options.source)
#                 if os.path.exists(temp):
#                     options.source = temp
#             
#         if options.test and not os.path.exists(options.test):
#             temp = os.path.join(options.usesharedfolder, "data", options.test)
#             if os.path.exists(temp):
#                 options.test = temp
#             else:
#                 temp = getDataFileInDocker(options.test)
#                 if os.path.exists(temp):
#                     options.test = temp
    if options.usesharedfolder:    
        options.folder = os.path.join(options.usesharedfolder, "workdir")
        options.targetfolder = os.path.join(options.usesharedfolder, "result")  # ic
    else:
        options.targetfolder = os.path.join(currentshare(options.caller), "result")
    
#     elif options.caller:
#         options.targetfolder = internalresult      
#     else:
#         options.targetfolder = externalresult
# #         racine = os.path.abspath(os.path.expanduser('~/nn1'))
# #         options.targetfolder = os.path.join(racine, "result")
    if options.debug & nnu.DEMO_DEBUG_FOLDER:
        doprint('options.usesharedfolder', options.usesharedfolder, file=sys.stderr)
        doprint('options.externpath', options.externpath, file=sys.stderr)
        doprint('options.targetfolder', options.targetfolder, file=sys.stderr)
            
#  adaptation de actionstr en fonction de origin 
    if os.path.exists(options.origin):
        if options.actionstr == "analyse,preproc":
            options.actionstr = 'usage'
        updateOptionsFromGmx(options, options.origin)
#     options.argumentfile = argumentfile
    if options.actionstr:
        options.action = options.actionstr.split(',')
    elif options.action and isinstance(options.action, text_type):
        options.action = [options.action]
    options.action = [val.lower() for val in options.action]
    if 'usage' in options.action:
        options.action = ['usage']
        if not options.origin:
            toprint = "options error\n\tUsage mode needs a model file (ORIGIN)!"
            errorcode = nnu.OPTION_STOP
            return options, toprint, errorcode
        if not os.path.exists(options.origin):
            toprint = "options error\n\tmodel file '{0}' cannot be found!".format(options.origin)
            errorcode = nnu.OPTION_STOP
            return options, toprint, errorcode
        
# adaptation du type de fichier de sortie            
    options.copytype = options.copytype.lower()
    options.copyexcel = options.copytype.startswith('xls') #in ['xls', 'xlsx', 'xlsm']
    options.copytxtcsv = options.copytype[:3] in ['csv', 'txt']
    if options.copytxtcsv:
        if len(options.copytype) > 3:
            options.copytype = options.copytype[:3]
            options.decimal = options.copytype[3:]
        if options.copytype == 'csv':
            options.sep = ";"
        elif options.copytype == 'txt':
            options.sep = "\t"

# detection des demandes simplifies de  "LICENSE", "VERSION", "HELP"         
    if  options.outfile is None:
        options.outfile = debugOutput(sys.stdout)
    if options.version:
        licfile = os.path.expanduser("~/version.txt")
        header = ["nn1 docker container modules versions:"]
        if os.path.exists(licfile):
            with open(licfile, 'r') as ff:
                ver = ff.read().rstrip()
            header.append("container : {0}".format(ver))
        lst = getVersionList(header)
        toprint = "\n\t".join(lst)
        errorcode |= nnu.OPTION_STOP
        return options, toprint, errorcode
    if options.license:
        licfile = os.path.expanduser("~/LICENSE")
        errorcode |= nnu.OPTION_DISPLAY_FILE_AND_STOP
        return options, licfile, errorcode
    if options.dhelp:
        helpfile = os.path.expanduser("~/help.txt")
        errorcode |= nnu.OPTION_DISPLAY_FILE_AND_STOP
        return options, helpfile, errorcode
    
# initialisation de actionlist, notrain, light
    actions = options.action
    options.actionlist = [val.lower() for val in actions]
    options.actionlist = _actionListMgmt(options.actionlist)
    options.notrain = not 'train' in options.actionlist
    options.light = os.path.exists(os.path.expanduser("~/hello.txt"))
#    if not options.light:
    # options.verbosexls = max(1, options.verbosexls)

# adaptation des test implicites
    if 'test' in options.actionlist and not options.test:
        options.test = options.source
        if options.testfields == "":
            options.testfields = options.datafields
        if not options.testrange and options.datarange:
            options.testrange = "TEST"
        
# ==============================
    if options.debug & nnu.DEMO_DEBUG_OPTIONS:
        print("usesharefolder", options.usesharedfolder)
        print("caller", options.caller)
        print("targetfolder", options.targetfolder)
        print("folder", options.folder)
    
    if options.compiler.lower() == "default":
        options.compiler = ""

    if options.decor:
        if options.decor == "[]":
            options.decor = []
        else:
            decordic = defaultdict((lambda: ""), basicDecorDict)     
            lst = []
            if not isinstance(options.decor, (list, tuple)):
                options.decor = options.decor.split(',')
            for val in options.decor:
                dval = decordic[val]
                val = dval if dval else val
                if val:
                    lst.append(val)
            if not "mode" in lst:
                lst.append("mode")
            options.decor = lst
    else:
        options.decor = []
    
    test = options.orderfield.strip()
    try:
        options.orderfield = C.CORRECTED_FIELD[test]
    except KeyError:
        options.orderfield = test

    index = None
    if not options.source and ('analyse' in options.actionlist) and not ('usage' in options.actionlist):
        if nofile:
            options.source, index = "", -1
        else:
            #forcedefault = options.yesyes
            candidates, searchdir = _getDataFileCandidates(searchbase=searchbases)
            options.source, index = _getDockerDataFile(options.source, options.caller,
                searchdir=searchdir, candidates=candidates, datakind='datafile', 
                defaultindex=0, doprint=doprint, forcedefault=options.yesyes )
        if (index is None or index < 0) and not nofile:
            errorcode  |= nnu.OPTION_STOP
            toprint = "no datafile: action stopped"
            return options, toprint, errorcode
        indextest = None
        if (index is not None and (index >= 0)) and not nofile:
            options.notest = not yesNoQuestion("Will you add a test ?", 'y', printMessage="Test", forceyes=options.yesyes, doprint=doprint)
            if not options.notest:
                if yesNoQuestion("Use the same file for test ?", 'y', forceyes=options.yesyes, doprint=doprint):  #forceyes=False, 
                    options.test = ""
                else:
                    options.test, indextest = _getDockerDataFile(options.test, options.caller,
                        searchdir=searchdir, candidates=candidates, datakind='testfile',
                        defaultindex=index, forcedefault=options.yesyes, doprint=doprint)
        if not options.notest and (indextest is not None) and (indextest == index):
            options.test = ""

    updateOptionsFromGmx(options, options.origin)
    
    options.root = CCharOnlys(os.path.splitext(os.path.basename(options.source))[0], extended=True) 
    if not options.root:
        options.root = CCharOnlys(os.path.splitext(os.path.basename(options.test))[0], extended=True)
    if not options.filetype:
        options.filetype = os.path.splitext(options.source)[1][1:]
    if not options.filetype:
        options.filetype = os.path.splitext(options.test)[1][1:]

    lst = options.source.split('|') 
    if len(lst) == 3:
        options.datafields = lst.pop(2)
    if len(lst) == 2:
        options.source = lst[0]
        if options.datarange:
            toprint = "{0}conflict of datarange. {1} will overwrite {2}\n".format(toprint, options.datarange, lst[1])
            errorcode |= nnu.OPTION_WARN_AND_ASK
        else:
            options.datarange = lst[1]
    
    if not options.datarange and options.filetype in multiRangesType:
        options.datarange = 'DATA'
    if not isinstance(options.datafields, list):
        options.datafields = getListFromStr(options.datafields)
    if not isinstance(options.testfields, list):
        options.testfields = getListFromStr(options.testfields)
    if not isinstance(options.grouplist, list):
        if options.grouplist:
            try:
                options.grouplist = [val.strip() for val in options.grouplist.split(',')]
            except ValueError:
                options.grouplist = [val.strip() for val in options.grouplist.split(';')]
            options.grouplist = [groupdict[val.lower()] for val in options.grouplist]
        else:
            options.grouplist = []
    if options.debug & nnu.DEMO_DEBUG_USAGE:
        print('initial grouplist :', options.grouplist)
    if options.groupcount and not isinstance(options.groupcount, list):
        try:
            options.groupcount = [int(val.strip()) for val in options.groupcount.split(',')]
        except ValueError:
            options.groupcount = [int(val.strip()) for val in options.groupcount.split(';')]
    if not options.testfields and not options.test and not options.testrange:
        options.notest = True
    cond = not hasattr(options, 'notest') or not options.notest
    if cond and not options.test and options.filetype in multiRangesType:
        options.test = options.source
    lst = options.test.split('|')  
    if len(lst) == 3:
        options.testfields = lst.pop(2)
    if len(lst) == 2:
        options.test = lst[0]
        if options.testrange:
            toprint = "{0}conflict of testrange. {1} will overwrite {2}\n".format(toprint, options.testrange, lst[1])
            errorcode |= nnu.OPTION_WARN_AND_ASK
        else:
            options.testrange = lst[1]
    if cond and not options.testrange and options.filetype in multiRangesType:
        options.testrange = 'TEST'
        
# on n'affiche pas le verbose < 5
    if (options.verbose >= 5) and (options.outfile is not None): # ic i
        toprint = "{0}verbose: {1}\n".format(toprint, options.verbose)
        toprint = "{0}sverbosexls: {1}\n".format(toprint, options.verbosexls)
#     if (options.verbose >= 5) and (outfile is not None):
        toprint = "{0}monal version: {1}\n".format(toprint, monalversion)
        toprint = "{0}nn1 version: {1}\n".format(toprint, nn1version)

    if not os.path.exists(options.folder):
        os.makedirs(options.folder, 0o777)
        mode = os.stat(options.folder)[0]
        os.chmod(options.folder, mode | stat.S_IWOTH | stat.S_IWGRP)
    if options.copytype: 
        try:
            targetfolder = options.targetfolder  #internalresult
            if not targetfolder:
                targetfolder = internalresult
            if not os.path.exists(targetfolder):
                os.makedirs(targetfolder, 0o777)
                mode = os.stat(targetfolder)[0]
                os.chmod(targetfolder, mode | stat.S_IWOTH | stat.S_IWGRP)
        except OSError:
            targetfolder = os.path.join(options.folder, 'result')
            if not os.path.exists(targetfolder):
                os.makedirs(targetfolder, 0o777)
                mode = os.stat(targetfolder)[0]
                os.chmod(targetfolder, mode | stat.S_IWOTH | stat.S_IWGRP)
        options.targetfolder = targetfolder
    if isInt(options.modelcreation):
        options.modelcreation = int(options.modelcreation)
    if not options.modelcreation:
        options.debug |= nnu.DEMO_USE_EXISTING
    
    if options.roundthreshold > 0:
        options.extraselect = "((DeltaRank = 0) AND (abs(RoundOffTest)< {0:g}))".format(options.roundthreshold)
        options.extraselectrelaxed = '(abs(RoundOffTest)<{0:g})'.format(options.roundthreshold)
    else:
        options.extraselect = "(DeltaRank = 0)"
        options.extraselectrelaxed = ''
    
    if options.savefolder and not os.path.isdir(options.savefolder):
        options.origin = options.savefolder
        folder = getcustomproperty(options.savefolder, 'Utilities', 'workingfolder')
        options.savefolder = folder

    if options.savefolder and not os.path.exists(options.savefolder):
        toprint = "bad savefolder option: %s\n"% options.savefolder
        errorcode |= nnu.OPTION_WARN_AND_STOP
        return options, toprint, errorcode
    
    if options.leveragethreshold <= 0:
        options.leveragethreshold = float('inf') 
        
    try:
        options.configstrMem = options.configstr
    except: pass
    
    options.multiseed = getListFromStr(options.seed)
    options.seed = options.multiseed[0]
    
    options.multihidden = getListFromStr(options.hidden)
    options.hidden = options.multihidden[0]   
    options.multihidden.sort()
    if len(options.multiseed) > 1:
        dirname = "{0}-multiseed".format(options.root)
        options.targetfolder = os.path.join(options.targetfolder, dirname)
        options.targetfolder = IndexFilename(options.targetfolder)
        if not os.path.exists(options.targetfolder):
            os.makedirs(options.targetfolder)

    options.multimaxrec = getListFromStr(options.maxrec)
    options.multimaxrec.sort(reverse=True)
    # traitement des valeurqs superieures a options.initcount
    excessvalues = [value for value in options.multimaxrec if value > options.initcount]
    while len(excessvalues):
        val = excessvalues.pop(0)        
        if len(excessvalues):
            options.multimaxrec.pop(0)
            toprint = "%sselected results %d is too high: ccanceled\n" % (toprint, val)
        else:
            options.multimaxrec[0] = options.initcount
            st = ','.join([str(val) for val in options.multimaxrec])
            toprint = "%sproposed selected results list: [%s]\n"%(toprint, st)
            errorcode |= nnu.OPTION_WARN_AND_ASK
    options.maxrec = options.multimaxrec[0]    

#     if not options.verbosexls:
#     # si options.verbosexls est 0, on n'affiche que le premier maxrec 
#         if (len(options.multimaxrec) > 1) and (options.verbose >= 3):
#             toprint = "%sselected results limited to %d only (only one value permitted)\n" % (toprint, options.multimaxrec[0])
#             errorcode |= nnu.OPTION_WARN_AND_ASK
#         options.multimaxrec = options.multimaxrec[:1]
    
    if options.debug & nnu.DEMO_DEBUG_OPTIONS:
        lst = ["options"]
        for val in dir(options):
            if not val.startswith("_"):
                value = options.__getattribute__(val)
                typ = type(value)
                name = typ.__name__
                lst.append("%s\t%s (%s)" % (val, value, name))
        toprint = "%s%s\ndone\n"%(toprint, "    \n".join(lst))
        errorcode |= nnu.OPTION_WARN_AND_STOP

    return options, toprint, errorcode
# end api_options

def saveOptionsToFile(options, forceAction="train,test", doprint=doNothing, forceYes=False, forceModel=False, rootTargetName="arguments"):
    """Write options namespace to a text file.
    
    Parameters:
     - options -> namespace to write down. 
         - The saving folder is read in options.savedir
     - doprint -> print function.
     - forceYes -> if set to False, the actual saving is submitted to accept question.
     - rootTargetName -> root of the written argument filename, without extension. The argument filename will be "options.root_rootfilename<_'x'>.txt", where 'x' is an index.
     
    Return:
     - options namespace eventually modified.
     - full name of the written argument file.
    """
# preparation de l'enregistrement du fichier d'arguments
    argdir = options.savedir
    if argdir.endswith('workdir'):
        argdir = argdir[:-len('workdir')] + 'argsdir'
        if not os.path.exists(argdir):
            os.makedirs(argdir, 0o777)
    argfilename = os.path.join(argdir, "{0}_{1}.txt".format(options.root, rootTargetName))
    argfilename = IndexFilename(argfilename, "_{0}")
    options.actionstr = forceAction
    options.yes = True
    lst = option2Arguments(options)
    if options.verbose >= 4:
        for st in lst:
            print(st)
    locargfilename = callerView(argfilename, options.caller, options.externpath)
#     if options.caller:
#         locargfilename = "~{0}".format(argfilename)
#     else:       
#         locargfilename = argfilename  
    mess = "Will you save the argument list in file '{0}'?".format(locargfilename) 
    if yesNoQuestion(mess, 'y', "", doprint=doprint, forceyes=forceYes):
        with open(argfilename, "wb") as ff:
            for val in lst:
                ff.write("{0}\n".format(val).encode())
        doprint("saved argument file: {0}".format(locargfilename))
        if options.caller:
            st = "cammand line:\n\tdocker run -it --rm -v ~/docker:{0} --env-file {1} <docker image>"
            message = st.format(internalshare, locargfilename)
            doprint(message)
        if forceModel:
            options.modelcreation = True
        options.yes = True
        options.actionstr = "train,test"
        options.actionlist = options.actionstr.split(',')
        return options, argfilename
    return options, ""

def sourceAnalysis(options=None, doprint=doNothing, forceAnalyse=True):
    """Data file analysis.
    
    Parameters:
     - options -> Namespace to write with the datasource informations and the user choices.
     - doprint -> print function.
     - forceAnalyse -> If True, a full datafile analysis is perdirmed.
    
    Return:
     - options -> options namespace.
     - modeldataframe -> object modelDataFrame for training data.
     - modeltestframe -> object modelDataFrame for testing data.
    """

    doprint("running source files analysis")
    newtrain = False
    if options is None:
#         options = _initiate_options([])
        options, _, _ = api_options([])
        options.yes = False
    oldyes = options.yes
    if 'analyse' in options.actionlist:
        if not options.yesyes:
            options.yes = False
    if hasattr(options, 'source') and options.source:
        sourceloc = os.path.basename(options.source)
        doprint("\tsourcefile : {0}".format(sourceloc))
    if hasattr(options, 'test') and options.test:
        testloc = os.path.basename(options.test)
        doprint("\ttestfile : {0}".format(testloc))     
    if hasattr(options, 'origin') and options.origin:
        originloc = os.path.basename(options.origin)
        doprint("\toriginfile : {0}".format(originloc))    
    newtrain = 'train' in options.actionlist
    try:
        res = analyseDataFiles(options, doprint=doprint, forceAnalyse=forceAnalyse)
    except Exception as err:
        doprint("Error in NN1_API.analyseDataFiles")
        doprint(err)
        return None

    options.source = res[0] 
    options.filetype = res[1] 
    options.datarange = res[2]
    options.datatitles = res[3]
    options.datafields = res[4]
    options.test = res[5]
    options.testfiletype = res[6] 
    options.testrange = res[7]
    options.testtitles = res[8]
    options.testfields = res[9]
    options.configstr = res[10]
    options.grouplist = res[11]
    options.classif = res[12]
    config = res[13]
    _, options.central, _, _, _ = ReverseConfigString(options.configstr) 
    if (options.verbose >= 4):
        if options.datarange:
            doprint("\tdata range : {0}".format(options.datarange))
        else:
            doprint("\tdata fields")
        for group, titles in zip(options.grouplist, options.datatitles):
            fieldlist = ", ".join(titles)
            doprint("\t{0} -> {1}".format(group, fieldlist))
        if options.test:
            doprint("\ttest source : {0}".format(options.test))
            if options.testrange:
                doprint("\ttest range : {0}".format(options.testrange))
            else:
                doprint("\ttest fields")
            
    dirname, basename = os.path.split(options.folder)
    if basename != "workdir":
        savedir = os.path.join(options.folder, "workdir", options.root)
        options.savedir = os.path.join(options.folder, "workdir")
    else:
        savedir = os.path.join(options.folder, options.root)
        options.savedir = options.folder
    #savedir = os.path.join(options.folder, options.root)
    savedir = IndexFilename(savedir, "_{0}", last=not newtrain, docreate=True)
    if not os.path.exists(savedir):
        os.makedirs(savedir)
    filest = "{0}.cfg".format(options.root)
    cfgfile = os.path.join(savedir, filest)
    if newtrain:
        inputnames = ""
        if len(options.datatitles) and isinstance(options.datatitles[0], list):
            if 'inputs' in options.grouplist:
                index = options.grouplist.index('inputs')
                inputnames = ",".join(options.datatitles[index])
        outputnames = ""
        if len(options.datatitles) and isinstance(options.datatitles[0], list):
            if 'outputs' in options.grouplist:
                index = options.grouplist.index('outputs')
                outputnames = ",".join(options.datatitles[index])
        config.set("model", "outputname", outputnames)
        config.set("model", "inputnames", inputnames)
        config._filename = cfgfile  
        config.save()
        
    options.savefolder = savedir
    options.cfgfile = cfgfile
    
    if 'analyse' in options.actionlist:
        mess = "Will you check the training parameters ?"
        if options.yesyes:
            accept = False
        else:
            accept = yesNoQuestion(mess, 'y', doprint=doprint)
        if accept:
            try:
                res = _CheckTrainingParameters(options, doprint=doprint)
            except Exception as err:
                doprint(err)
                return None
        if 'loo' in options.actionlist:
            mess = "Will you check the Leave-One-Out parameters ?"
            accept = yesNoQuestion(mess, 'y', doprint=doprint)
            if accept:
                try:
                    res = _CheckLOOParameters(options, doprint=doprint)
                except Exception as err:
                    doprint(err)
                    return None
        if 'bootstrap' in options.actionlist:
            mess = "Will you check the Bootstrap parameters ?"
            accept = yesNoQuestion(mess, 'y', doprint=doprint)
            if accept:
                try:
                    res = _CheckBSParameters(options, doprint=doprint)
                except Exception as err:
                    doprint(err)
                    return None
    options.yes = oldyes
    doprint('source files analysis done')
    return options  #, modeldataframe, modeltestframe 

def _CheckInputTypes(filename, datarange, datafields, datagroups, 
        filetype='', forcedefault=False, doprint=doNothing):
    drop = ('identifiers' in datagroups) and not datagroups.index('identifiers') and (len(datafields[0]) == 1)
    modeldataframe, indexdict = get_modelDataframe(filename, datarange, datafields, 
        datagroups, filetype=filetype, drop=drop, withindexdict=True)  #, sheetname=None, indexfield=-1, drop=False)
    typeDict = modeldataframe.dtypes
    outlist = list(modeldataframe.fieldsOf('outputs'))
    for outputname in outlist:
        if typeDict[outputname] == str:
            raise nn1Error("Outputs must be numeric")
    inlist = list(modeldataframe.fieldsOf('inputs'))
    for inputname in inlist:
        if typeDict[inputname] == str:
            mess = "Input field '{0}' has string type.\n\tWill you switch to Graphmachine engine ?".format(inputname)
            accept = yesNoQuestion(mess, 'y', forceyes=forcedefault, doprint=doprint)
            if accept:
                ind = inlist.index(inputname)
                inindex = modeldataframe.dataGroups.index('inputs')
                inputindex = modeldataframe.dataFields[inindex][ind]
                smilst = [modeldataframe.dataFields[inindex].pop(ind)]
                decal =  int('identifiers' in modeldataframe.dataGroups)
                modeldataframe.dataGroups.insert(decal, 'smiles')
                modeldataframe.dataFields.insert(decal, smilst)
    removelist = [ind for ind, fields in enumerate(modeldataframe.dataFields) if not len(fields)]
    removelist.sort(reverse=True)
    for ind in removelist:
        modeldataframe.dataFields.pop(ind)
        modeldataframe.dataGroups.pop(ind)
    modeldataframe._datacounts = [len(val) for val in modeldataframe.dataFields]
    if drop:
        datafields = [datafields[datagroups.index('identifiers')]] + [[indexdict[val] for val in lst] for lst in modeldataframe.dataFields]
        datagroups = ['identifiers'] + modeldataframe.dataGroups
    else:
        datagroups = modeldataframe.dataGroups
        datafields = [[indexdict[val] for val in lst] for lst in modeldataframe.dataFields]

    return modeldataframe, datafields, datagroups

def _chooseInOccurList(title, originlist, connectdico=None, doprint=doNothing):
    if connectdico is None:
        connectdico = defaultdict(lambda:-1)
    lst = ["{0} occurences and connections".format(title)]
    for element, occ, occmol in originlist:
        defaultval = connectdico[element]
        lst.append( "{0}\t{1}\tin {2} examples. Proposed {3}".format(element, occ, occmol, defaultval))
    st = "\n\t".join(lst)
    doprint(st)
    mess = "Do you accept these values ?"
    accept = yesNoQuestion(mess, 'y', doprint=doprint)
    if not accept:
        for element, occ, occmol in originlist:
            defaultval = connectdico[element]
            prefix = "Enter the connectivity\n"
            message = "\t{0}\t{1}\tin {2} examples ->".format(element, occ, occmol)
            printMessage = "connectivity"
            connectdico[element] = valueQuestion(mess=message, prefix=prefix, default=defaultval, 
                printMessage=printMessage, doprint=doprint)
    return connectdico

def analyseGraphmachineData(dataFrame, config, configstr="", acceptdefault=False, classif=False, doprint=doNothing):
    """Analyse data file for gathering chemical informations for graphmachine configuration string creation.
    
    Parameters:
     - dataFrame -> modelDataFrame of training data.
     - configstr -> proposed origin configuration string.
     - classif -> classification odel.
     - doprint -> print function.
     
    Return the configuration string. 
    """
    if not dataFrame:
        return ""
    _connect, central, _classif, _hidden, _fullH = ReverseConfigString(configstr, False)
    if central == []:
        central = 3
    res = analysebase(iterator=dataFrame.iterData(), fullH=_fullH, titlesfirst=True)
    atomlst = res[0]    # res
    grademax = res[1]   # grade
    isomax = res[2]     # iso
    chiralmax = res[3]  # chiral
    occurs = res[4]     # occurs
    occurmols = res[5]  # occurmols
    cycles = res[6]     # cycles
#    delimiter = res[7]
    datacount = res[8]      # data count
    baseindex = res[9]      # liste des molecules
    smilist = res[10]   # smilist
    valueslist = res[11] # list des valeurs de propriete
    moldetails = res[12]   # liste des dictionnaires des occurences par molecules
    titles = res[13]   # titles
    baselimit = res[14] #  baselimit
    chargelist = res[15]
    errorlist = res[16]  # errorlist
    limited = (baselimit is not None)
    
    dict3 = {
        "grademax": grademax,
        "isomax": isomax,
        "chiralmax": chiralmax,
        "cyclemax": cycles,
        "atoms": ",".join(atomlst),
        'fullhydrogen': _fullH,     
        }
    if len(chargelist):
        dict3["chargemin"] = chargelist[0]
        dict3["chargemax"] = chargelist[1]
    if limited:
        dict3["limited"] = str(baselimit)    
    config.updatesection("general", **dict3)

    
    dict1 = {}
    dict2 = {}
    for key, value in occurs.items():
        st = "%s, %s" %(value, occurmols[key])
        if key in atomlst: 
            dict1[key] = st
        else: # key in occurmols
            dict2[key2long(key)] = st
    for key, value in cycles.items():
        dict2[cyclekey(key+1, True)] = value  #cycles[key]
    config.updatesection("occurence", **dict1)
    config.updatesection( "exoccurence", **dict2)


    
    structlist = [value for value in occurs if not value in atomlst]
    structlist.sort()
    
    if not acceptdefault:
        prefix = "Root atom is chosen among the 'N' first atoms in {0}\n".format(atomlst)
        message = "\tPlease choose the number 'N' of root candidates"
        printMessage = 'N'
        central = valueQuestion(mess=message, prefix=prefix, default=central, errorValue=0,
                    printMessage=printMessage, doprint=doprint)
        
    if central:
        config.set('model', 'central', str(central))
    else:
        config.set('model', 'central', '')
    atomconnectdico = defaultdict(lambda:-1)
    for atom in atomlst:
        if atom in _connect.keys():
            atomconnectdico[atom] = _connect[atom]
        else:
            atomconnectdico[atom] = defaultConnect[atom]
    # ce dico a ete initialise avec des valeurs par default precedentes venant de _connect
    structconnectdico = defaultdict(lambda:-1)
    for struct in structlist:
        if struct in _connect.keys():
            structconnectdico[struct] = _connect[struct]
        else:
            structconnectdico[struct] = defaultConnect[struct]
    for key, val in _connect.items():
        if key == 'grds':
            for keyloc in structconnectdico.keys():
                if keyloc.startswith('grd'):
                    structconnectdico[keyloc] = val
    # ce dico a ete initialise avec des valeurs par default precedentes venant de _connect
    atomOccurList = [(atom, occurs[atom], occurmols[atom]) for atom in atomlst]
    structOccurList = [(struct, occurs[struct], occurmols[struct]) for struct in structlist]
    if not acceptdefault:
        atomconnectdico = _chooseInOccurList("Atoms", atomOccurList, atomconnectdico, doprint=doprint)
        structconnectdico = _chooseInOccurList("Structures", structOccurList, structconnectdico, doprint=doprint)
    
    atomconnectdico.update(structconnectdico)
    configstr = getConfigString(atomlst, atomconnectdico, central, classif, hidden=None, short=True)
    
    return configstr, config

def analyseDataFiles(options=None, commonFieldExclude=True, 
        forceAnalyse=False, doprint=doNothing):
    """Analyse the content of data files.
    
    Parameters:
     - options -> prposerd parameter namespace.
         Following attributes must exists:
          - options.source -> data file.
     - commonFieldExclude -> Exclusion of common fields in data groups. 
     - forceAnalyse -> force the full analyse of data file.
     - doprint -> print function.
     
    Return a tuple containing:
     - dataframe -> modelDataFrame of training data
     - testframe -> modelDataTrain of test data
     - sourcefile -> data training file
     - datafiletype -> type of data file
     - datarange -> range of train data
     - datafields -> list of used fields in data file/ range for train
     - datafieldlist ->  list of fields index in each data group for train
     - testfile ->  data testing file
     - testfiletype ->  type of test file 
     - testrange ->  range of train test
     - testfields ->  list of used fields in data file/ range for test
     - testfieldlist -> list of fields index in each data group for test
     - configstr -> configuration string for graphmachine model construction.
     - grouplist ->  list of data groups. A group may be one of ['indentifiers', 'smiles', 'inputs', 'outputs']
     - classif ->  classification model.
    """
    
    if options is None:
        options, _, _ = api_options([])
        options.yes = False
    accept = True
    if 'analyse' in options.actionlist:
        if not forceAnalyse:
            mess = "Will you check the data file(s) ?"
            accept = yesNoQuestion(mess, 'y', doprint=doprint)
        if accept:
            if options.grouplist == []:
                options.grouplist = ['inputs', 'outputs']
                options.groupount = [[0], [0]]
            forceyes=not accept or options.yesyes
            options.grouplist = _checkGroupList(options.grouplist, 
                forceyes=forceyes, doprint=doprint, verbose=options.verbose)
    if len(options.groupcount) != len(options.grouplist):
        options.groupcount = [1 for _ in options.grouplist]
    sourcefile = options.source
    filetype = options.filetype
    datarange = options.datarange
    datafieldlist = options.datafields
    if datafieldlist is None or not len(datafieldlist):
        datafieldlist = []
        options.datafields = []
    testfile = options.test
    testrange = options.testrange
    testfieldlist = options.testfields
    forcedefault = options.yes or not accept
    grouplist = options.grouplist
    groupCount = options.groupcount 
    verbose = options.verbose 
    commonFieldExclude = True,
     
    assert sourcefile or testfile, "Need a data file"
    
    commonfiles = testfile == ""
    testfiletype = ""
    testfields = []
    dotrain = 'train' in options.actionlist
    if 'analyse' in options.actionlist or (dotrain and (datafieldlist == [] or datafieldlist is None)):  #not filetype or 
        filetype, datarange, datafields, ranges = _checkFile(
            sourcefile, 
            datarange=datarange,
            datafor='train', 
            indexdefault=0, 
            forcedefault=forcedefault, 
            doprint=doprint, 
            verbose=verbose)
        assert datafields is not None, "File {0} has been refused. \nCannot proceeed".format(sourcefile)
        assert len(datafields), "Need data in file {0}".format(sourcefile)
        fieldprompt = "Please choose {0} values by their indexes comma separated for the {1} group in the followings:"
        fieldprompt1 = "Please choose the index of the {0} field"
        if 'analyse' in options.actionlist:
            try:
                defaultlist = datafieldlist.copy()
            except:
                defaultlist = datafieldlist
        if not 'analyse' in options.actionlist or not len(defaultlist):
            defaultlist = []
            value = 0
            for countval in groupCount:
                lloc = []
                for _ in range(countval):
                    lloc.append(value)
                    value += 1
                defaultlist.append(lloc)
        printMessage = "Chosen fields"
        datafieldlist = chooseFields(fieldprompt, fieldprompt1, datafields, 
            grouplist=grouplist, groupCount=groupCount, defaultlist=defaultlist,
            mutulaExclude=commonFieldExclude, printMessage=printMessage,
            doprint=doprint, forcedefault=forcedefault, acceptnull=True, 
            verbose=verbose, debug=1)   
        idlst = datafieldlist[0]
        if len(idlst) == 1 and idlst[0] < 0:
            grouplist.pop(0)
            groupCount.pop(0)
            datafieldlist.pop(0)
        datafiletype = filetype
        while [] in datafieldlist:
            indexnull = datafieldlist.index([])
            grouplist.pop(indexnull)
            datafieldlist.pop(indexnull)
    elif sourcefile:
        if not os.path.exists(sourcefile):
            mess = "Sorry, cannot find data source {0}".format(sourcefile)
            raise nn1Error(mess)
        datafiletype = filetype
        datafields = gettitles(sourcefile, filetype, datarange=datarange, datafmt=options.datafields)
    else:
        datafiletype = filetype
        datafields = []
    if len(datafieldlist):
        options.groupcount = [len(val) for val in datafieldlist]

#=======================================================================#    
#    input type checking 
    dataframe, datafieldlist, grouplist = _CheckInputTypes(sourcefile,
        datarange, datafieldlist, grouplist, datafiletype, 
        forcedefault=options.yesyes ,doprint=doprint)
# here, a 'smiles' item may be included in grouplist, and 
# 'inputs' item may have desappeared
#=======================================================================# 
   
    testframe = None        
    mode = 'gm' if 'smiles' in grouplist else 'nn'
    options.mode = mode
    if verbose >= 3:
        doprint("\tmode {0}".format(mode))
    if dotrain or not 'analyse' in options.actionlist:
        classif = options.classif
    else:
        mess = "Is this model a classification model ?"
        classif = yesNoQuestion(mess, 'n', forceyes=options.yesyes, doprint=doprint)
    configstr = options.configstr
    config = defaultCfg(mode)
    if mode == 'gm':  #'smiles' in grouplist:
        acceptdefault = not forceAnalyse or options.yesyes 
        configstr, _ = analyseGraphmachineData(dataframe, config=config, configstr=configstr, acceptdefault=acceptdefault, classif=classif, doprint=doprint)
        if configstr:
            doprint("\tconfigstr: {0}".format(configstr))
        else:
            doprint("\tconfigstr empty")
        
    notest = options.filetype in simpleRangeType and not testfile
    notest = notest or(options.filetype in multiRangesType and not testrange)
    if not notest:
        if 'analyse' in options.actionlist or testfieldlist == []:
            if commonfiles:
                testfileloc = sourcefile
            else:
                testfileloc = testfile   
            testfiletype, testrange, testfields, ranges = _checkFile(
                testfileloc, 
                datafor='test', 
                datarange=testrange,
                indexdefault=1, 
                forcedefault=forcedefault, 
                doprint=doprint, 
                verbose=verbose) 
        dotestanalysis = 'analyse' in options.actionlist and (testfields != datafields)
        commonstruct = len(datafields) and ((testfieldlist == []) or (testfieldlist is None) or (testfields == datafields))
        if not dotestanalysis and commonstruct:
            mess = "Data and test files seems to have the same structrure. Will you set the sames field choice ?"
            res = yesNoQuestion(mess, 'y', "", doprint=doprint, forceyes=forcedefault, verbose=verbose)  
            if res:
                testfieldlist = datafieldlist
            else:    
                dotestanalysis = True
        if dotestanalysis:
            fieldprompt = "Please choose {0} values by their indexes comma separated for the {1} group in the followings:"
            fieldprompt1 = "Please choose the index of the {0} field"
            defaultlist = []
            value = 0
            for countval in groupCount:
                lloc = []
                for _ in range(countval):
                    lloc.append(value)
                    value += 1
                defaultlist.append(lloc)
            printMessage = "Chosen fields"
            testfieldlist = chooseFields(fieldprompt, fieldprompt1, testfields, 
                grouplist=grouplist, groupCount=groupCount, defaultlist=defaultlist,
                mutulaExclude=commonFieldExclude, printMessage=printMessage,
                doprint=doprint, forcedefault=forcedefault, acceptnull=True,
                verbose=verbose)   
            testfiletype = filetype
            if not len(testfields):
                testfields = gettitles(testfile, filetype, datarange=testrange, datafmt=None)
        else:
            testfiletype = filetype
            testfields = gettitles(testfile, filetype, datarange=testrange, datafmt=options.testfields)
                             
    return sourcefile, datafiletype, datarange, datafields, datafieldlist, testfile, testfiletype, testrange, testfields, testfieldlist, configstr, grouplist, classif, config
            
def preProcAction(options, doprint=doNothing, forceYes=False):
    """Input data selection with the random probe method.
    
    Arguments:
     - options -> options namespace.
     - doprint -> print function
     - forceYes -> force the action without question.
     
    Following the action, the data selection in options may be changed
    """
    if 'preproc' in options.actionlist and 'inputs' in options.grouplist:
        mess = "do you want to perform the input selection"
        accept = yesNoQuestion(mess, default="y", doprint=doprint, forceyes=forceYes)
        if accept:           
            getout = False
            doprint("running nn1_api input selection")
            curyes = options.yes
            options.yes = False
            try:
                preprocres = _preprocAction(options, doprint)
            except Exception as err:
                doprint(err)
                getout = True
            finally:
                options.yes = curyes
                doprint('nn1_api input selection done')
            if getout:
                return -1
    return 0 

def doPreProc(options, Inputs, Outputs, inputTitles, probecount=-1, doprint=doNothing):
    activepreprocs = preproctypelist[1:]
    if options.preproctype in activepreprocs:
        preprocindex = activepreprocs.index(options.preproctype)
        order2 = preprocindex > 0
        crossprod = preprocindex == 2
        if probecount < 0:
            probecount = valueQuestion("Enter number of random probe tests", 
                options.probecount, "Probe tests", doprint=doprint, 
                verbose=options.verbose, forcedefault=options.yes)  #, prefix=prefix)
        options.probecount = probecount
        with Progress_tqdm(options.probecount, desc='random probe', nobar=options.verbose<nobarverb) as update:
            Pertinence = randomProbe(Inputs, Outputs, options.probecount, 
                seed=options.seed, randomSource='shuffle', order2=order2, 
                crossprod=crossprod, progress=update)  #options.preproctype"
        vecind = orderedIndexes(Pertinence)
#         doprint()
        message = "Probe results with {0} selection ({1}):".format(options.preproctype, options.probecount)
        printresult(options, message, vecind, Pertinence, 
            inputTitles=inputTitles, printTitles=True, doprint=doprint)
    else:
        doprint("Random probe analysis failed")
        Pertinence = vecind = None
    return Pertinence, vecind

def _DoPreProcToExcel(options, verbose=3, doprint=doNothing):
    curpreproc = options.preproctype
    source = DataSource(options.source, options.filetype, options.datarange, 
        datafmt=options.datafields, titles=True)
    res = source.getDataArrays(options.datarange, fieldsList=options.datafields, 
        titles=True, transpose=True)
    inputTitles = res[-1][options.grouplist.index('inputs')]
    Inputs = res[options.grouplist.index('inputs')]
    Outputs = res[options.grouplist.index('outputs')] 
    perts = []
    vecinds = []
    probecount = valueQuestion("Enter number of random probe tests", 
        options.probecount, "Probe tests", doprint=doprint, 
        verbose=options.verbose, forcedefault=options.yes)  #, prefix=prefix)
    for preproctype in preproctypelist[1:]:
        options.preproctype = preproctype
        Pertinence, vecind = doPreProc(options, Inputs, Outputs, inputTitles, probecount=probecount)  #, doprint=doprint)
        if Pertinence is None:
            break
        perts.append(Pertinence)
        vecinds.append(vecind)  
    pertinenceFrame = DataFrame(perts, columns=inputTitles, index=preprocnamelist[1:])
    
    
    print(pertinenceFrame)
    
    
    options.preproctype = curpreproc
    return pertinenceFrame, vecinds
     
def _preprocAction(options, doprint=doNothing):
    pertinenceFrame, vecinds = _DoPreProcToExcel(options, doprint=doprint)
    inputTitles = list(pertinenceFrame.columns)
    dopreproc = options.preproctype in preproctypelist[1:] 
    res = False 
    while dopreproc:
        preproctype = preproctypelist.index(options.preproctype)  
        mess = "Choose the preprocessor action by its index."
        printMessage = "Preprocessor action"
        choicelist = ["No_Action", "linear random probe", "simple 2nd order random probe", "full 2nd order random probe"]
        preproctype, index = choiceQuestion(mess, choicelist,
            preproctype, printMessage, returnType=str, doprint=doprint, 
            extradisplay=True, listindexed=True, forcedefault=options.yes,
            fulloutput=True)
        if not index: 
            return None
        options.preproctype = preproctypelist[index]
        Pertinence = pertinenceFrame.ix[index-1]
        vecind = vecinds[index-1]
        
        if Pertinence is None:
            return -1
        doselect = True
        while doselect:
            try:
                options.pertinencethreshold = valueQuestion("Enter the input selection pertinence", 
                    options.pertinencethreshold, "Input selection pertinence", 
                    forcedefault=options.yes, doprint=doprint, 
                    verbose=options.verbose, returnType=float) 
            except Exception as err:
                doprint(str(err))
                return 1
            res = False            
            if (0.0 < options.pertinencethreshold < 1.0):
                choices = [val < options.pertinencethreshold for val in Pertinence]
                if any(choices):
                    #doprint()
                    mess = "Following inputs may be removed:"
                    printresult(options, mess, vecind, Pertinence, 
                        inputTitles=inputTitles, choices=choices, printTitles=False, 
                        doprint=doprint)
                    mess = "Do you accept?"
                    selcount = 0
                    for choice in choices:
                        if not choice:
                            selcount += 1
                    OKChoice = vecind[:selcount]
                    OKChoice.sort()
                    OKTitles = [inputTitles[val] for val in OKChoice]
                    st = ", ".join(OKTitles)
                    printMessage = ""  
                    res = yesNoQuestion(mess, default='Y', printMessage=printMessage, 
                        forceyes=options.yes, verbose=options.verbose, 
                        doprint=doprint)
                    if res:
                        inputindex = options.grouplist.index('inputs')
                        inputini = options.datafields[inputindex]
                        testinputini = options.testfields[inputindex]
                        Inputs = [val for ind, val in enumerate(inputini) if ind in OKChoice]
                        options.groupcount[inputindex] = len(Inputs)
                        options.datafields[inputindex] = Inputs
                        if options.test:
                            Inputs = [val for ind, val in enumerate(testinputini) if ind in OKChoice]
                            options.testfields[inputindex] = Inputs
                if res:
                    doselect = False
                else:
                    mess = "will you try a new selection threshold ? "
                    doselect = yesNoQuestion(mess, default='n', printMessage="", 
                            forceyes=options.yes, verbose=options.verbose, 
                            doprint=doprint)
        if res:
            dopreproc = False
            doprint("Inputs will be {0}".format(st))
        else:
            mess = "will you try a new pre-processing ? "
            dopreproc = yesNoQuestion(mess, default='y', printMessage="", 
                    forceyes=options.yes, verbose=options.verbose, 
                    doprint=doprint)
    if res:
        return options
    return None

# def _preprocActionOld(options, doprint=doNothing):
#     source = DataSource(options.source, options.filetype, options.datarange, 
#         datafmt=options.datafields, titles=True)
#     res = source.getDataArrays(options.datarange, fieldsList=options.datafields, 
#         titles=True, transpose=True)
#     inputTitles = res[-1][options.grouplist.index('inputs')]
#     Inputs = res[options.grouplist.index('inputs')]
#     Outputs = res[options.grouplist.index('outputs')] 
#     dopreproc = options.preproctype in preproctypelist[1:] 
#     res = False 
#     while dopreproc:
#         preproctype = preproctypelist.index(options.preproctype)  
#         mess = "Choose the preprocessor action by its index."
#         printMessage = "Preprocessor action"
#         choicelist = ["No_Action", "linear random probe", "simple 2nd order random probe", "full 2nd order random probe"]
#         preproctype, index = choiceQuestion(mess, choicelist,
#             preproctype, printMessage, returnType=str, doprint=doprint, 
#             extradisplay=True, listindexed=True, forcedefault=options.yes,
#             fulloutput=True)
#         if not index: 
#             return None
#         options.preproctype = preproctypelist[index]
#         Pertinence, vecind = doPreProc(options, Inputs, Outputs, inputTitles, doprint=doprint)
#         if Pertinence is None:
#             return -1
#         doselect = True
#         while doselect:
#             try:
#                 options.pertinencethreshold = valueQuestion("Enter the input selection pertinence", 
#                     options.pertinencethreshold, "Input selection pertinence", 
#                     forcedefault=options.yes, doprint=doprint, 
#                     verbose=options.verbose, returnType=float) 
#             except Exception as err:
#                 doprint(str(err))
#                 return 1
#             res = False            
#             if (0.0 < options.pertinencethreshold < 1.0):
#                 choices = [val < options.pertinencethreshold for val in Pertinence]
#                 if any(choices):
#                     #doprint()
#                     mess = "Following inputs may be removed:"
#                     printresult(options, mess, vecind, Pertinence, 
#                         inputTitles=inputTitles, choices=choices, printTitles=False, 
#                         doprint=doprint)
#                     mess = "Do you accept?"
#                     selcount = 0
#                     for choice in choices:
#                         if not choice:
#                             selcount += 1
#                     OKChoice = vecind[:selcount]
#                     OKChoice.sort()
#                     OKTitles = [inputTitles[val] for val in OKChoice]
#                     st = ", ".join(OKTitles)
#                     printMessage = ""  
#                     res = yesNoQuestion(mess, default='Y', printMessage=printMessage, 
#                         forceyes=options.yes, verbose=options.verbose, 
#                         doprint=doprint)
#                     if res:
#                         inputindex = options.grouplist.index('inputs')
#                         inputini = options.datafields[inputindex]
#                         testinputini = options.testfields[inputindex]
#                         Inputs = [val for ind, val in enumerate(inputini) if ind in OKChoice]
#                         options.groupcount[inputindex] = len(Inputs)
#                         options.datafields[inputindex] = Inputs
#                         if options.test:
#                             Inputs = [val for ind, val in enumerate(testinputini) if ind in OKChoice]
#                             options.testfields[inputindex] = Inputs
#                 if res:
#                     doselect = False
#                 else:
#                     mess = "will you try a new selection threshold ? "
#                     doselect = yesNoQuestion(mess, default='n', printMessage="", 
#                             forceyes=options.yes, verbose=options.verbose, 
#                             doprint=doprint)
#         if res:
#             dopreproc = False
#             doprint("Inputs will be {0}".format(st))
#         else:
#             mess = "will you try a new pre-processing ? "
#             dopreproc = yesNoQuestion(mess, default='y', printMessage="", 
#                     forceyes=options.yes, verbose=options.verbose, 
#                     doprint=doprint)
#     if res:
#         return options
#     return None

def trainAction(options=None, sourceframe=None, hidden=3, moduledir="", 
        modulename="", initcount=12, multimodel=10, epochs=150, seed=0, 
        initstddev=0.1, extraselect="", trainstyle=0, savedir="", 
        tablename='trainingResult', verbose=3, doprint=doNothing, 
        forceyes=False, prefix="\t", resultFrameList=[]):
    """Training action
    """

    resultFrame = None
    
    if options is not None:
        hidden = options.hidden
        moduledir = options.folder
        initcount = options.initcount
        multimodel = options.maxrec
        epochs = options.epochs
        seed = options.seed
        initstddev = options.initstddev
        forceyes = options.yes
        verbose = options.verbose
        root = options.root
        cfgfile = options.cfgfile
        tablename = options.tableName
        pptyname = options.propertyName
        pptyunit = options.propertyunit
         
    config = defaultedConfigParser(cfgfile)
    if not config.has_section("projects"):
        config.add_section("projects")
    config.set("projects", "0", root)
    if not config.has_section(root):
        config.add_section(root)
    config.set(root, 'tablename', tablename)
    config.set(root, "propertyname", pptyname)
    config.set(root, "propertyunit", pptyunit)

    if hidden < 0:
        hiddendefault = - hidden
        hidden = valueQuestion("Enter the number of hidden neurons", 
            hiddendefault, "Hidden neurons", doprint=doprint, 
            verbose=verbose, forcedefault=forceyes, prefix=prefix)

    
    initcount = valueQuestion("Enter the number of training run", initcount, 
        "Training count", forcedefault=forceyes, prefix=prefix, doprint=doprint,
        verbose=verbose)  
    options.initcount = initcount  
    epochs = valueQuestion("Enter the number epochs per run", epochs, 
        "Epochs", forcedefault=forceyes, prefix=prefix, doprint=doprint,
        verbose=verbose)    
    options.epochs = epochs
    if (seed > 0):
        doprint("{0}seed : {1}".format(prefix, seed))
    if verbose > 3:
        doprint("{0}random std-dev : {1:g}".format(prefix, initstddev))
    config.set("model", "hidden", str(hidden))
    config.set("training", "initcount", str(initcount))
    config.set("training", "epochs", str(epochs))
    config.set("training", "seed", str(seed))
    config.save()
    
    # chargement ou creation du module d'apprentissage
    
    baselen = sourceframe.shape[0]
    updatetuple = (baselen, 'sourcing module', verbose<3, sys.stdout)
    doprinttime = False
    debug = 0
    lib, fullmodname, modeldata = createTrainingModule(options, sourceframe,
        config=config, outputdir=savedir, compiler="", 
        progress=updatetuple, keeptemp=options.keeptemp,
        verbose=verbose, doprinttime=doprinttime, debug=debug)
    if lib is None:
        raise nn1Error("training module '{0}' cannot be created or loaded".format(localize_file(options, fullmodname)))
    
    config.set(root, "module", fullmodname)

    options.paramCount = lib.paramCount
    datacount, datasize = lib.loadTrainingDataFromFrame(sourceframe)
    options.modulename = fullmodname
    options.datacount = datacount
    options.datasize = datasize

    if doprint:
        if (verbose >= 4):
            doprint(lib.repr('long'))
        if (verbose == 3):
            doprint(lib.repr('short'))
    
    
    multimodel = min(multimodel, initcount)
    if not savedir:
        try:
            savedir = options.savefolder
        except: pass
    if not savedir:
        savedir = os.path.join(moduledir, root)
        savedir = IndexFilename(savedir, "_{0}", docreate=True)
        if not os.path.exists(savedir):
            os.makedirs(savedir)
    options.savefolder = savedir
    sqlFileName = os.path.join(savedir, "%s.db"% root)
    config.set(root, 'sqlfilename', sqlFileName)
    if os.path.exists(sqlFileName):
        os.remove(sqlFileName)
    dbConnect = sqlconnect(sqlFileName, detect_types=sqlPARSE_DECLTYPES)
    options.sqlFileName = sqlFileName
    traindata = C.TRAIN_DATA
    try:
        createTableDb(dbConnect, tablename, traindata)
    except sqlOperationalError:
        cleanTableDb(dbConnect, tablename)
    propertyname = pptyname if pptyname else options.datatitles[options.grouplist.index('outputs')][0]
    
    config = setTrainingActionParams(config, propertyname, initcount=initcount, 
        initstddev=initstddev, epochs=epochs, resultdir=moduledir, seed=seed, 
        criterion=C.TCR_PRESS)
    config.save(cfgfile)
    
    with Progress_tqdm(initcount, desc='training', nobar=verbose<nobarverb) as update:
        onEndIterloc = ft.partial(onEndIter, progress=update, dowrite=None)
        onReturnResultloc = ft.partial(onReturnResult, dbConnect, tablename)
        nfail = lib.multitrain(
                initcount=initcount,
                epochs=epochs,
                weightsstddev=initstddev, 
                seed=seed,
                onEndIter=onEndIterloc,
                onReturnResult=onReturnResultloc,
                trainstyle=trainstyle)
    bestweight = _getBestWeights(dbConnect, options)
    
    lib.weights = bestweight
    resultFrame = getResultFrame(dbConnect, options=options)
    if not resultFrame.shape[0]:
        st = "All the trained models are degenerated. The full rank condition is now relaxed."
        doprint(st)
        options.extraselect = options.extraselectrelaxed
        resultFrame = getResultFrame(dbConnect, options=options)
    if "ID" in resultFrame.columns:
        resultFrame.set_index("ID", inplace=True)
    countmax = min(multimodel, resultFrame.shape[0])
    if countmax < multimodel:
        st = "Number of selected models has been reduced from {0} downto {1}.".format(multimodel, countmax)
        doprint(st)
    resultFrameList.append((options.hidden, resultFrame.copy(), lib.paramCount)) 
    
    # sauvegarde dans un fichier d'utilisation gmx ou nnx
    gm0 = getModelDict(config, lib, version=nn1version)  #, modeldata)
    indexlist = resultFrame.index[:countmax] 
    gmxfile, dico = getGmxFileAndSave(dbConnect, gm0, config, indexlist, options, "nn1", full=True)  # ici gmx
    
    if len(options.multimaxrec):
        for val in options.multimaxrec[1:]:
            newfile, mess = getReducedSelectedGmx(dico, val, gmxfile)
            if mess:
                doprint(mess)

    confResult = config
    confResult.set(root, "mcount", "%d" % countmax)
    confResult.set(root, "gmxfile", gmxfile)
    confResult.set(root, "sqlfilename", sqlFileName)
    confResult.set(root, "tablename", tablename)
    confResult.set(root, "paramnames", ";".join(list(lib.parameterNames)))
    if lib.inputCount:
        confResult.set(root, 'inputnorm', ";".join([str(val) for val in lib.inputnorm]))
    confResult.set(root, 'outputnorm', ";".join([str(val) for val in lib.outputnorm]))
    confResult.set(root, 'paramcount', "{0}".format(lib.paramCount))
    confResult.set(root, 'baselen', "{0}".format(lib.modelCount))
                   
    confResult.save()
    dbConnect.close()
    if verbose and doprint:
        if nfail:
            doprint("%s fails\n"% nfail)
        elif verbose >= 2:
            formatters = {'RoundoffTest': roundoff_format(), 'ID': IDformat()}
            doprint("%s best training results:"% countmax)
            columns = [resultFrame.index.name] + list(resultFrame.columns)
            st = resultFrame.to_string(
                float_format=float_format(verbose), index_names=True, 
                formatters=formatters)
            lst = st.split("\n")
            lst1 = lst[1].strip()
            ll1 = len(lst1)
            tst = lst[0][:ll1]
            OK = True
            for c in tst:
                OK = OK and (c==" ")
            if OK:
                st = lst1 + lst[0][ll1:]
                lst[0] = st
                lst.pop(1)
                st = "\n".join(lst)
            doprint(st)
    
     
    return fullmodname, resultFrame, cfgfile, gmxfile, dico, sqlFileName

def looTrainAction(options, model, sourceFrame, resultFrame, doprint=doNothing):
    seed = options.seed
    looTrainFrame = None
    lootrainfailed = False
    traintype = 'loo'
    sqlFileName = options.sqlFileName
    resfile = options.cfgfile
    config = defaultedConfigParser(options.cfgfile)
    try:
        res = _apiDemoLOO(config, options, sourceFrame, resultFrame, model, seed)  #, resfile)
        model, looTrainFrame, gmxfileLOO, sqlLOOFileName, IDlist, \
            LOOResultTrainColumn = res
    except:
        lootrainfailed = True   
        raise
    return looTrainFrame, lootrainfailed, LOOResultTrainColumn +1, gmxfileLOO, sqlLOOFileName          

def bsTrainAction(options, module, doprint=doNothing):
    doprint("bsTrainAction not yet implemented")
    pass

def usageAction(options=None, modelfile=None, data=""):
    """Usage of a model resulting from training action.
    
     - modelfile -> filename in format NML, NNX or GMX, holding the training results.
     - options -> namespace holding datas. Following fields may be used:
         - options.source -> data file
         - options.datarange -> range of data for excel type files
         - options.grouplist -> list of data types from ('identifiers', 'smiles', 'inputs', 'outputs')
         - options.groupcount -> list of datacount in each type
         - options.datafields -> list of data indexes per data types.
     - data -> single line data in format "identifier;[smiles;][inputs[,followings inputs comma separated];][target output]".
         - identifier may be a name or "_" if no known name
         - smiles is a string data in SMILES format, representing a chemical molecule. Optional
         - inputs is a comma separated list of float values. Optional. At least one of smiles and inputs field must be present.
         - output is a float value representing the target value. Optional
    """
    res = prepareUsageAction(options, modelfile, data)
    datamode, data, source, mode, inputnames, outputnames, nh, datalist,\
        grouplist, extracount, xml, dsize = res
    if 'identifiers' in grouplist:
        idName = datalist[grouplist.index('identifiers')][0]
    else:
        idName = None
    target = None
    ni = len(inputnames)
    no = len(outputnames)
    model = None
    inputs = None
    smile = ""
    if datamode == "single":
        id = ""
        if 'identifiers' in grouplist:
            id = datalist[grouplist.index('identifiers')][0]
        if options.debug & nnu.DEMO_DEBUG_USAGE:
            print("datalist", datalist)
            print("grouplist", grouplist)
        if mode == 'nn':
            if xml:
                model = Driver(xml)
            else:
                modeldata = ModelData(input=ni, output=no, hidden=nh, name=id)
                model = Driver(modeldata=modeldata)
            model.loadFromDict(source)
            if not idName:
                idName = model.mainModel.name
            inputs = np.asarray([float(val) for val in datalist[grouplist.index('inputs')]])
            if 'targets' in grouplist:
                target = float(datalist[grouplist.index('targets')][0])
            
            outs, levs, stds, students = model.transferEx(inputs, withCI=True)  
        elif mode == 'gm': 
            if 'inputs' in grouplist:
                inputs = np.asarray([float(val) for val in datalist[grouplist.index('inputs')]])
            if 'targets' in grouplist:
                target = float(datalist[grouplist.index('targets')][0])
            smile = datalist[grouplist.index('smiles')][0]
            
            model = createGMModel(source, id, smile, truename="")  #savedir=None, 
            if not idName:
                idName = model.modelName
            outs, levs, stds, students = model.transferEx(inputs, 
                withCI=True) 
        
        usageFrame = manageSingleUsage(options, model, outs, levs, stds, 
            students)
        finishSingleUsage(usageFrame, idName, smile, inputs, target, 
            inputnames, outputnames[0], options.leveragethreshold, 
            dsize, options.verbose, options.verbosexls)
        
    elif datamode == 'multi':
        datafile = options.test
        testrange = options.testrange
        fieldlist = options.testfields
        drop = ('identifiers' in grouplist) and not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
        dataFrame = get_modelDataframe(datafile, testrange, fieldlist, grouplist, drop=drop) 
        cols = ['mean_{0}'.format(outputnames[0]), 'SCount']
        fieldlist = ["estimated", 'min', 'max', "leverage"]
        cols = getSimpleExtraFields(extracount, outputnames[0], fieldlist, cols)
        fullCol = list(dataFrame.columns) + cols
        newcol = OrderedDict({val: float('nan') for val in cols})  
        resultFrame = dataFrame.assign(**newcol)  
        resultFrame = resultFrame[fullCol] 
        modelNames = []
        outslist = []
        levslist = []
        icslist = []
        stdslist = []
        studentslist = []
        count = 0
        id = "NNUsage"
        if mode == 'nn':
            if xml:
                model = Driver(xml)
            else:
                modeldata = ModelData(input=ni, output=no, hidden=nh, name=id)
                model = Driver(modeldata=modeldata)
            model.loadFromDict(source)
            modelName = model.mainModel.name            
            with Progress_tqdm(dataFrame.shape[0], desc="computing model", nobar=options.verbose<nobarverb) as update:
                for ID, row in dataFrame.iterrows():
                    count += 1
                    update(count)
                    if ni:
                        inputs = row[:ni]
                    smile = row[ni]
                    target = row[ni+1:]
                    res = model.transferEx(inputs, withCI=True)  #original=True
                    outs, levs, stds, students = res
                    ics = [np.sqrt(lev)*studentvar*stddev for lev, stddev, studentvar in zip(levs, stds, students)]
                    outslist.append(outs)
                    levslist.append(levs)
                    stdslist.append(stds)
                    studentslist.append(students)
                    icslist.append(ics)
                    
        elif mode == 'gm':
            with Progress_tqdm(dataFrame.shape[0], desc="computing model", nobar=options.verbose<nobarverb) as update:
                for ID, row in dataFrame.iterrows():
                    count += 1
                    update(count)
                    if ni:
                        inputs = row[:ni]
                    smile = row[ni]
                    target = row[ni+1:]
                    model = createGMModel(source, ID, smile, truename="")  #savedir=None, 
                    modelNames.append(model.modelName)
                    res = model.transferEx(inputs, withCI=True)  #original=True
                    outs, levs, stds, students = res
                    ics = [np.sqrt(lev)*studentvar*stddev for lev, stddev, studentvar in zip(levs, stds, students)]
                    outslist.append(outs)
                    levslist.append(levs)
                    stdslist.append(stds)
                    studentslist.append(students)
                    icslist.append(ics)
    #                extraVLOOs = model.extraVLOO if model.extraVLOO is not None else [np.nan for _ in range(outs)]
                
        resultFrame = finishMultiUsage(resultFrame, outputnames[0], fieldlist, modelNames, outslist, levslist, icslist, options.leveragethreshold, options.verbosexls)
        _, excelwriter = createExcelwriter(options, resultFrame, "Usage")  #, with index='identifiers' in options.grouplist)
        usageSheet = getSheet(excelwriter, "Usage")
        excelresult = CloseExcelwriter(options, excelwriter, Sheet=usageSheet)
        dispExcelresult = callerView(excelresult, options.caller, options.externpath)
        mess = "results in file {0}".format(dispExcelresult)
        print(mess)

def testAction(options, testframe, datacount, IDlist, 
        cfgfile, gmxfile, dico, doprint=doNothing):
    if not options.test:
        options.test = options.source
    sourceroot = options.root 
    if not options.savedir:
        options.savedir = os.path.join(options.folder, sourceroot)
        options.savedir = IndexFilename(options.savedir, "_{0}", True)
    if hasattr(options, 'libuse'):
        libuse = options.libuse
    else:
        libuse = True
    if libuse:
        options.debug &= (~nnu.DEMO_USAGE_NO_LIB)
    else:
        options.debug |= nnu.DEMO_USAGE_NO_LIB
    if options.notrain and len(options.multimaxrec):
        for val in options.multimaxrec[1:]:
            newfile, mess = getReducedSelectedGmx(dico, val, gmxfile)
            if mess:
                doprint(mess)

    res = apiTestResults(cfgfile, datacount=datacount, 
        modeltestframe=testframe, IDlist=IDlist, options= options)                                      
    testFrame, _, _, resultColumn, gmxfile = res  #, testinputColumns

    return testFrame, resultColumn #,testinputColumns

def looTestAction(options, testframe, resfile, IDlist, doprint=doNothing):  #tempres, ex celwriter, sheetname, apiTrainRes, 
    mode = options.mode
    looTestFrame = None
    traintype = 'loo'
    if not options.test:
        options.test = options.source
    sourceroot = options.root  #os.path.splitext(os.path.basename(options.source))[0]
    if not options.savedir:
        options.savedir = os.path.join(options.folder, sourceroot)
        options.savedir = IndexFilename(options.savedir, "_%d", True)
    if not resfile:
        resfile = os.path.join(options.savedir, "train.cfg")

    if options.libuse:
        options.debug &= (~nnu.DEMO_USAGE_NO_LIB)
    else:
        options.debug |= nnu.DEMO_USAGE_NO_LIB
    
    datacount = testframe.shape[0]
    looTestFrame, patternlist, _, _, _ = \
        apiTestResults(resfile, datacount, testframe, IDlist, 
        options, leveragethreshold=options.leveragethreshold,
        #dataframe=looTrainFrame, 
        extraselect="", orderfield="PRESS",  traintype='loo', 
        todoprint=True);  
                                            
    return looTestFrame

def bsTestAction(options, doprint=doNothing):
    doprint("bsTestAction not yet implemented")
    pass

def getVersionList(header=[]):
    lst = ["{0} : {1}".format(key, value) for key, value in dicoversion.items()]
    if len(header):
        header.extend(lst)
        lst = header
    return lst

def getVersions(dec=""):
    lst = ["{0} : {1}".format(key, value) for key, value in dicoversion.items()]
    return "\n{0}".format(dec).join(lst)
    
if __name__ == '__main__':

    print(getVersions())
#     opt, toprint, errorcode = api _options([], nofile=True)
#     lst = option2Arguments(opt)
#     for val in lst:
#         print (val)
#     print('')

#     api_o ptions(help=True)
#     doprint = toPrint(sys.stdout)
#     options = sourceAnalysis(doprint=doprint)  #, modelDataFrame, modelTestFrame
#     saveOptionsToFile(options, doprint=doprint)
#     res = analyseDataFiles(forceAnalyse=True, doprint=doprint)
#     modelDataFrame, sourcefile, datafiletype, datarange, datafields, datafieldlist, testfile, testfiletype, testrange, testfields, testfieldlist, configstr, config = res
#     print(options)
    print('Done')